@@ -32,23 +32,17 @@ from yellowdog_client.model import (
     WorkRequirementStatus,
 )
 
-from yd_commands.args import ARGS_PARSER
-from yd_commands.config import (
-    CONFIG_FILE_DIR,
-    ConfigWorkRequirement,
-    generate_id,
-    link_entity,
-    load_config_work_requirement,
-    update_config_work_requirement,
-)
-from yd_commands.config_keys import *
+from yd_commands.config_types import ConfigWorkRequirement
 from yd_commands.csv_data import (
     csv_expand_toml_tasks,
     load_json_file_with_csv_task_expansion,
     load_jsonnet_file_with_csv_task_expansion,
     load_toml_file_with_csv_task_expansion,
 )
+from yd_commands.follow_utils import follow_events
+from yd_commands.id_utils import YDIDType
 from yd_commands.interactive import confirmed
+from yd_commands.load_config import CONFIG_FILE_DIR, load_config_work_requirement
 from yd_commands.printing import (
     WorkRequirementSnapshot,
     print_error,
@@ -56,10 +50,14 @@ from yd_commands.printing import (
     print_log,
     print_numbered_strings,
 )
+from yd_commands.property_names import *
+from yd_commands.settings import NAMESPACE_PREFIX_SEPARATOR
 from yd_commands.submit_utils import (
     UploadedFiles,
     format_yd_name,
     generate_task_input_list,
+    pause_between_batches,
+    update_config_work_requirement,
 )
 from yd_commands.type_check import (
     check_bool,
@@ -70,6 +68,7 @@ from yd_commands.type_check import (
     check_str,
 )
 from yd_commands.upload_utils import unique_upload_pathname
+from yd_commands.utils import generate_id, link_entity
 from yd_commands.validate_properties import validate_properties
 from yd_commands.variables import (
     L_TASK_COUNT,
@@ -84,9 +83,9 @@ from yd_commands.variables import (
     load_json_file_with_variable_substitutions,
     load_jsonnet_file_with_variable_substitutions,
     load_toml_file_with_variable_substitutions,
-    process_variable_substitutions,
+    process_variable_substitutions_in_dict_insitu,
 )
-from yd_commands.wrapper import CLIENT, CONFIG_COMMON, main_wrapper
+from yd_commands.wrapper import ARGS_PARSER, CLIENT, CONFIG_COMMON, main_wrapper
 
 # Import the Work Requirement configuration from the TOML file
 CONFIG_WR: ConfigWorkRequirement = load_config_work_requirement()
@@ -170,18 +169,22 @@ def main():
             )
         validate_properties(wr_data, "Work Requirement JSON")
         submit_work_requirement(
-            directory_to_upload_from=dirname(wr_data_file),
+            directory_to_upload_from=(
+                dirname(wr_data_file)
+                if ARGS_PARSER.content_path is None
+                else ARGS_PARSER.content_path
+            ),
             wr_data=wr_data,
         )
 
     else:
-        task_count = (
-            CONFIG_WR.task_count
-            if ARGS_PARSER.task_count is None
-            else ARGS_PARSER.task_count
-        )
         submit_work_requirement(
-            directory_to_upload_from=CONFIG_FILE_DIR, task_count=task_count
+            directory_to_upload_from=(
+                CONFIG_FILE_DIR
+                if ARGS_PARSER.content_path is None
+                else ARGS_PARSER.content_path
+            ),
+            task_count=CONFIG_WR.task_count,
         )
 
     if ARGS_PARSER.dry_run:
@@ -218,7 +221,7 @@ def submit_work_requirement(
     )
     # Lazy substitution of the Work Requirement name, now it's defined
     add_substitutions(subs={L_WR_NAME: ID})
-    process_variable_substitutions(wr_data)
+    process_variable_substitutions_in_dict_insitu(wr_data)
     CONFIG_WR = update_config_work_requirement(CONFIG_WR)
 
     # Handle any files that need to be uploaded
@@ -227,7 +230,13 @@ def submit_work_requirement(
 
     # Ensure we're in the correct directory for uploads
     if directory_to_upload_from != "":
-        chdir(directory_to_upload_from)
+        try:
+            chdir(directory_to_upload_from)
+        except Exception as e:
+            raise Exception(
+                "Unable to switch to content directory"
+                f" '{directory_to_upload_from}': {e}"
+            )
 
     # Flatten upload paths?
     flatten_upload_paths = check_bool(
@@ -459,21 +468,29 @@ def add_tasks_to_task_group(
         wr_data[TASK_GROUPS][tg_number][TASKS] = [{}]
         num_tasks = 1
 
-    # If 'taskCount' is set at the Task Group level, and there is only one Task,
-    # create 'taskCount' copies of the Task.
-    # Note: NOT inherited from the TOML or WR levels.
-    task_group_task_count = wr_data[TASK_GROUPS][tg_number].get(TASK_COUNT, None)
+    # If the 'taskCount' property is set, and there is only one Task
+    # in the Task Group, create 'taskCount' duplicates of the Task.
+    task_group_task_count = check_int(
+        wr_data[TASK_GROUPS][tg_number].get(
+            TASK_COUNT, wr_data.get(TASK_COUNT, CONFIG_WR.task_count)
+        )
+    )
     if task_group_task_count is not None:
-        if num_tasks == 1:
+        if num_tasks == 1 and task_group_task_count > 1:
             # Expand the number of Tasks to match the specified Task count
-            for _ in range(1, int(task_group_task_count)):
+            print_log(
+                f"Expanding number of Tasks in Task Group '{task_group.name}' to"
+                f" 'taskCount={task_group_task_count}' Tasks"
+            )
+            for _ in range(1, task_group_task_count):
                 wr_data[TASK_GROUPS][tg_number][TASKS].append(
                     deepcopy(wr_data[TASK_GROUPS][tg_number][TASKS][0])
                 )
-        else:
+        elif task_group_task_count > 1:
             print_log(
-                f"Warning: Task Group '{task_group.name}' contains {num_tasks} "
-                "Tasks: ignoring 'taskCount'"
+                f"Note: Task Group '{task_group.name}' already contains"
+                f" {num_tasks} Tasks: ignoring expansion using 'taskCount ="
+                f" {int(task_group_task_count)}'"
             )
 
     num_task_groups = len(wr_data[TASK_GROUPS])
@@ -499,12 +516,14 @@ def add_tasks_to_task_group(
 
     # Iterate through batches
     for batch_number in range(num_task_batches):
-        if ARGS_PARSER.pause_between_batches and batch_number > 0:
-            print_log(
-                f"Pausing before submitting batch number {batch_number + 1}. "
-                "Press enter to continue:"
+        if ARGS_PARSER.pause_between_batches is not None:
+            pause_between_batches(
+                task_batch_size=TASK_BATCH_SIZE,
+                batch_number=batch_number,
+                num_tasks=num_tasks,
+                first_batch=batch_number == 0,
             )
-            input()
+
         # Iterate through tasks in the batch
         for task_number in range(
             (TASK_BATCH_SIZE * batch_number),
@@ -526,7 +545,7 @@ def add_tasks_to_task_group(
             add_substitution_overwrite(
                 L_TASK_NUMBER, formatted_number_str(task_number, num_tasks)
             )
-            process_variable_substitutions(task)
+            process_variable_substitutions_in_dict_insitu(task)
 
             executable = check_str(
                 task.get(
@@ -549,10 +568,10 @@ def add_tasks_to_task_group(
             # Set up lists of files to input, verify
             input_files_list = check_list(
                 task.get(
-                    INPUT_FILES,
+                    INPUTS_REQUIRED,
                     task_group_data.get(
-                        INPUT_FILES,
-                        wr_data.get(INPUT_FILES, CONFIG_WR.input_files),
+                        INPUTS_REQUIRED,
+                        wr_data.get(INPUTS_REQUIRED, CONFIG_WR.inputs_required),
                     ),
                 )
             )
@@ -578,7 +597,7 @@ def add_tasks_to_task_group(
                     INPUTS_OPTIONAL,
                     task_group_data.get(
                         INPUTS_OPTIONAL,
-                        wr_data.get(INPUTS_OPTIONAL, CONFIG_WR.optional_inputs),
+                        wr_data.get(INPUTS_OPTIONAL, CONFIG_WR.inputs_optional),
                     ),
                 )
             )
@@ -647,10 +666,10 @@ def add_tasks_to_task_group(
                 TaskOutput.from_worker_directory(file_pattern=file, required=False)
                 for file in check_list(
                     task.get(
-                        OUTPUT_FILES,
+                        OUTPUTS_OPTIONAL,
                         task_group_data.get(
-                            OUTPUT_FILES,
-                            wr_data.get(OUTPUT_FILES, CONFIG_WR.output_files),
+                            OUTPUTS_OPTIONAL,
+                            wr_data.get(OUTPUTS_OPTIONAL, CONFIG_WR.outputs_optional),
                         ),
                     )
                 )
@@ -661,20 +680,38 @@ def add_tasks_to_task_group(
                 TaskOutput.from_worker_directory(file_pattern=file, required=True)
                 for file in check_list(
                     task.get(
-                        OUTPUT_FILES_REQUIRED,
+                        OUTPUTS_REQUIRED,
                         task_group_data.get(
-                            OUTPUT_FILES_REQUIRED,
-                            wr_data.get(
-                                OUTPUT_FILES_REQUIRED, CONFIG_WR.output_files_required
-                            ),
+                            OUTPUTS_REQUIRED,
+                            wr_data.get(OUTPUTS_REQUIRED, CONFIG_WR.outputs_required),
                         ),
                     )
                 )
             ]
 
-            # Set 'alwaysUpload' to true for all outputs
-            for task_output in outputs:
-                task_output.alwaysUpload = True
+            # Add the contents of the 'outputsOther' property
+            for output_other in check_list(
+                task.get(
+                    OUTPUTS_OTHER,
+                    task_group_data.get(
+                        OUTPUTS_OTHER,
+                        wr_data.get(OUTPUTS_OTHER, CONFIG_WR.outputs_other),
+                    ),
+                )
+            ):
+                check_dict(output_other)
+                try:
+                    outputs.append(
+                        TaskOutput.from_directory(
+                            directory_name=output_other[DIRECTORY_NAME],
+                            file_pattern=output_other[FILE_PATTERN],
+                            required=output_other[REQUIRED],
+                        )
+                    )
+                except KeyError as e:
+                    raise Exception(
+                        f"Missing field in property '{OUTPUTS_OTHER}' ({e})"
+                    )
 
             # Add TaskOutput to 'outputs'?
             if check_bool(
@@ -688,6 +725,18 @@ def add_tasks_to_task_group(
             ):
                 outputs.append(TaskOutput.from_task_process())
 
+            always_upload = task.get(
+                ALWAYS_UPLOAD,
+                task_group_data.get(
+                    ALWAYS_UPLOAD,
+                    wr_data.get(ALWAYS_UPLOAD, CONFIG_WR.always_upload),
+                ),
+            )
+
+            # Set 'alwaysUpload' for all outputs
+            for task_output in outputs:
+                task_output.alwaysUpload = always_upload
+
             # If there's no task type in the task definition, AND
             # there's only one task type at the task group level,
             # use that task type
@@ -704,7 +753,8 @@ def add_tasks_to_task_group(
                     wr_data=wr_data,
                     task_group_data=task_group_data,
                     task_data=task,
-                    name=task_name,
+                    task_name=task_name,
+                    tg_name=task_group.name,
                     task_type=task_type,
                     executable=executable,
                     args=arguments_list,
@@ -780,9 +830,10 @@ def get_task_data_property(
                 return f.read()
 
 
-def follow_progress(work_requirement: WorkRequirement) -> None:
+def follow_progress_old(work_requirement: WorkRequirement) -> None:
     """
-    Follow and report the progress of a Work Requirement
+    Follow and report the progress of a Work Requirement.
+    Deprecated temporarily due to problems with Python 3.10+.
     """
     listener = DelegatedSubscriptionEventListener(on_update=on_update)
     CLIENT.work_client.add_work_requirement_listener(work_requirement, listener)
@@ -795,6 +846,15 @@ def follow_progress(work_requirement: WorkRequirement) -> None:
         print_log(f"Work Requirement did not complete: {work_requirement.status}")
 
 
+def follow_progress(work_requirement: WorkRequirement) -> None:
+    """
+    Follow and report the progress of a Work Requirement.
+    Replacement for the SDK version above.
+    """
+    print_log("Following Work Requirement event stream")
+    follow_events(work_requirement.id, YDIDType.WORK_REQ)
+
+
 def on_update(work_req: WorkRequirement):
     """
     Print status messages on Work Requirement update
@@ -842,7 +902,7 @@ def deduplicate_inputs(task_inputs: List[TaskInput]) -> List[TaskInput]:
             )
             print_log(
                 f"Removing '{task_input.verification}' duplicate:"
-                f" '{namespace}::{task_input.objectNamePattern}'"
+                f" '{namespace}{NAMESPACE_PREFIX_SEPARATOR}{task_input.objectNamePattern}'"
             )
 
     return deduplicated_task_inputs
@@ -942,7 +1002,8 @@ def create_task(
     wr_data: Dict,
     task_group_data: Dict,
     task_data: Dict,
-    name: str,
+    task_name: str,
+    tg_name: str,
     task_type: str,
     executable: str,
     args: List[str],
@@ -956,6 +1017,8 @@ def create_task(
     Create a Task object, handling special processing for specific Task Types.
     """
 
+    env_copy = deepcopy(env)  # Copy the environment property to prevent overwriting
+
     def _make_task(flatten_input_paths: FlattenPath) -> Task:
         """
         Helper function to create the Task object.
@@ -965,11 +1028,11 @@ def create_task(
             flatten_input_paths = None
 
         return Task(
-            name=name,
+            name=task_name,
             taskType=task_type,
             arguments=args,
             inputs=inputs,
-            environment=env,
+            environment=env_copy,
             outputs=outputs,
             flattenInputPaths=flatten_input_paths,
             taskData=task_data_property,
@@ -991,6 +1054,16 @@ def create_task(
     ):
         flatten_input_paths = FlattenPath.FILE_NAME_ONLY
 
+    # Environment variable names
+    yd_task_name_str = "YD_TASK_NAME"
+    yd_tg_name_str = "YD_TASK_GROUP_NAME"
+    yd_wr_name_str = "YD_WORK_REQUIREMENT_NAME"
+
+    # Add names to the environment as a convenience
+    env_copy[yd_task_name_str] = task_name
+    env_copy[yd_tg_name_str] = tg_name
+    env_copy[yd_wr_name_str] = ID
+
     # Special processing for Bash, Python & PowerShell tasks if the 'executable'
     # property is set. The script is uploaded if this hasn't already been done,
     # and added to the list of required files.
@@ -1031,7 +1104,7 @@ def create_task(
         return _make_task(flatten_input_paths)
 
     # Special processing for Docker tasks if the 'executable property is set.
-    # Sets up the '-e' environment strings and the DockerHub username and
+    # Sets up the '--env' environment strings and the DockerHub username and
     # password if specified.
     elif task_type == "docker":
         if executable is None:
@@ -1047,13 +1120,15 @@ def create_task(
                 ),
             )
         )
-        # 'TASK_NAME' env. var. is set for convenience
-        docker_env_string = f"--env TASK_NAME={name}"
+        # Names are set in the container env. for convenience
+        docker_env_list = ["--env", f"{yd_task_name_str}={task_name}"]
+        docker_env_list += ["--env", f"{yd_tg_name_str}={tg_name}"]
+        docker_env_list += ["--env", f"{yd_wr_name_str}={ID}"]
         if docker_env is not None:
             for key, value in docker_env.items():
-                docker_env_string += f" --env {key}={value}"
+                docker_env_list += ["--env", f"{key}={value}"]
 
-        args = [docker_env_string, executable] + args
+        args = docker_env_list + [executable] + args
 
         # Set up the environment used by the script to run Docker
         # Add the username and password, if present
@@ -1071,7 +1146,7 @@ def create_task(
                 wr_data.get(DOCKER_PASSWORD, CONFIG_WR.docker_password),
             ),
         )
-        env.update(
+        env_copy.update(
             {
                 "DOCKER_USERNAME": docker_username,
                 "DOCKER_PASSWORD": docker_password,
@@ -1110,7 +1185,7 @@ def submit_json_raw(wr_file: str):
     wr_data["name"] = format_yd_name(wr_data["name"])
     wr_name = wr_data["name"]
     add_substitutions(subs={L_WR_NAME: wr_name})
-    process_variable_substitutions(wr_data)
+    process_variable_substitutions_in_dict_insitu(wr_data)
 
     if ARGS_PARSER.dry_run:
         # This will show the results of any variable substitutions
@@ -1161,7 +1236,7 @@ def submit_json_raw(wr_file: str):
                         else input["namespace"]
                     )
                     verify_at_start_files.add(
-                        f"{namespace} :: {input['objectNamePattern']}"
+                        f"{namespace}{NAMESPACE_PREFIX_SEPARATOR}{input['objectNamePattern']}"
                     )
 
     # Warn about VERIFY_AT_START files & halt to allow upload or
